//  STOP.cpp

//  Copyright 2007 University of Pittsburgh.
//  U of Pittsburgh coding done by Brady Hunsaker and Mustafa Baz.
//
//  This file is part of STOP (Selection Tool for Optimization Parameters).
//
//  The contents of this file are subject to the Common Public License
//  1.0.  (the "License"); you may not use this file except in
//  compliance with the License. You should have received a copy of
//  the Common Public License along with STOP.
//
//  Software distributed under the License is distributed on an "AS
//  IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
//  implied. See the License for the specific language governing
//  rights and limitations under the License.
//
//  Alternatively, the contents of this file may be used under the
//  terms of the GNU General Public License Version 2 or later (the
//  "GPL"), in which case the provisions of the GPL are applicable
//  instead of those above. If you wish to allow use of your version
//  of this file only under the terms of the GPL, and not to allow
//  others to use your version of this file under the terms of the
//  CPL, indicate your decision by deleting the provisions above and
//  replace them with the notice and other provisions required by the
//  GPL. If you do not delete the provisions above, a recipient may
//  use your version of this file under the terms of either the CPL or
//  the GPL.
//

//  Dependencies:
//
//  FANN: If you wish to use neural networks, you need to have FANN 
//  (Fast Artificial Neural Network) installed on your system.  
//  Compilation will fail if it is not found.  FANN is 
//  licensed under the GNU LGPL.
//  If you do not wish to use FANN, compile with the NO_FANN flag.
//
//  R:  If you wish to use regression trees, then you must have R installed
//  on your system.  It is not linked to the code but called through system
//  calls, so there is no licensing issue.  
//  In particular, you need the rpart module.  If do not wish to use regression
//  trees then you do not need R.  If you do try to use them but R is not 
//  available then execution will fail during run-time.

#include <iostream>
#include <fstream>
#include <vector>
#include <queue>
#include <utility>
#include <string>

// Windows needs to use the getopt.h included in the STOP distribution
#ifdef WIN32
#include <time.h>
#define ULTRAGETOPT_REPLACE_GETOPT
extern "C" {
  #include "ultragetopt.h"
}
#else
// unix-like environments should be able to find getopt in the system
#include <getopt.h>
#include <sys/resource.h>
#endif

// Note: the code below for the Mersenne Twister pseudo-random 
// number generator has a modified BSD license.
// This file is redistributed with STOP.
#include "MersenneTwister.h"

#ifdef WIN32
// For now, assume FANN not available with Windows
#define NO_FANN
#endif

#ifndef NO_FANN
// You need to have FANN installed on your computer.
extern "C" {
  #include <doublefann.h>
}
#endif

#include "test_selection.hpp"

// Only one interface file should be included
#include "interface_glpk.hpp"
//#include "interface_cplex.hpp"
//#include "interface_cplex_extended.hpp"
//#include "interface_cplex_10_gap.hpp"
//#include "interface_cbc.hpp"

using namespace std;

#define BIG_DOUBLE 1.0e10;

//  One trial is determined by a vector of integers 
//  of size num_parameters.  
//  The i-th integer is in the domain 0..(par_levels[i]-1).
//  This file isn't concerned about what the parameters represent.  Just
//  think of them as vectors of integers.

// These global variables are used to convert between the encoding mentioned 
// above and the encoding the neural net uses, which is a little different 
// (so that it only uses binary input values).
vector<int> input_starts;
int num_inputs = 0;
int leftover;


// Utility functions defined below main function
void print_help_message ( void );
double run_test_instances ( STOP_interface & interface, 
			    vector<int> setting );
double time_now( void );
int exhaustive_index ( vector<int> setting, const int num_parameters, 
		       const vector<int> & par_levels );
vector<int> convert_params_to_input( const vector<int> & nn_inputs, 
				     const int num_parameters, 
				     const vector<int> & par_levels );
vector<int> convert_input_to_params( const vector<int> & params, 
				     const int num_parameters,
				     const vector<int> & par_levels );

int
main (int argc, char ** argv)
{
  ////////////////////////////////////////////////////
  // Instantiate the interface and extract needed values.
  // We must do this before parsing options, since some options
  // change these values.
  STOP_interface interface;

  const int num_parameters = interface.get_num_parameters();
  const vector<int> par_levels = interface.get_par_levels();
  const vector<int> default_settings = interface.get_default_settings();
  const vector<string> par_name = interface.get_par_name();
  const vector<vector<string> > par_level = interface.get_par_level();

  int num_initial_tests = interface.get_num_initial_tests();
  // number of settings for initial tests
  int num_repetitions = interface.get_num_repetitions();
  // number of repetitions(default is 1)
  int num_additional_tests = interface.get_num_additional_tests();
  // number of additional settings to test after machine learning

  int num_regression_tree_branchings = 2;
  // Number of branchings to check if using the regression tree.
  // Default is two, but may be decreased if the regression tree does
  // not create two branchings.


  /////////////////////////////////////
  // Common variables

  // Random number generator; uses the Mersenne Twister code
  MTRand randgen;  // this uses a random seed; an option below changes the seed
  
  
  ///////////////////////////////////////////
  // Parse options using "getopt" utility

  // After all the options, a single argument gives a file with 
  //  filenames of instances to test

  // To add an option that sets a flag, add a variable here
  int regression_tree_int = 0;
  int neural_net_int = 0;
  int naive_int = 0;
  int random_int = 0;
  int our_heuristic_int = 0;
  int pairwise_int = 0;
  int double_pairwise_int = 0;
  int triplewise_int = 0;
  int final_advice = 0;

  // For a long option (more than a letter), add a 4-tuple to this structure.
  // To set a single integer variable, model it on one of the lines below.
  // For anything more complicated see documentation for getopt_long.
  const struct option longopts[] = 
    {
      // Learning strategies
      { "regression-tree", no_argument, &regression_tree_int, 1 },
      { "neural-net", no_argument, &neural_net_int, 1 },
      { "naive", no_argument, &naive_int, 1 },
      // How to design trials
      { "random", no_argument, &random_int, 1 },
      { "heuristic", no_argument, &our_heuristic_int, 1 },
      { "pairwise", no_argument, &pairwise_int, 1 },
      { "double-pairwise", no_argument, &double_pairwise_int, 1 },
      { "number-initial-trials", required_argument, 0, 'n' },
      { "number-replicates", required_argument, 0, 'r' },
      { "number-additional-trials", required_argument, 0, 'a' },
      // Time limit 
      { "time-limit", required_argument, 0, 't' },
      // Specify output file or recovery file
      { "output", required_argument, 0, 'o' },
      { "recover", required_argument, 0, 'c' },
      { "advice", no_argument, &final_advice , 1 },
      // Set a random seed
      { "random-seed", required_argument, 0, 's' },
      ///// For development and testing:
      { "exhaustive", required_argument, 0, 'e' },
      { "input-file", required_argument, 0, 'i' },
      { 0, 0, 0, 0}
    };

  // For a single letter option, add the letter to the string below and
  // put a 'case' statement in the switch below for your letter.
  const char shortopts [] = "hn:os:a:r:";

  // Lots of boolean variables
  bool use_regression_tree;
  bool use_neural_net;
  bool use_naive;
  bool use_random;
  bool use_our_heuristic;
  bool use_pairwise;
  bool use_double_pairwise;
  bool use_triplewise;

  bool use_exhaustive = false;
  bool use_from_exhaustive_file = false;

  char * exh_filename;
  ofstream exh_output;
  ifstream exh_input;

  // Temporary handling of legacy variables for neural net input
  bool from_file = false;
  char * filename = "data.tmp";
  //  if (from_file)  filename = argv[1];


  int indexptr;  // index of the long opt
  int ret; // return code
  do
    {
      ret = getopt_long(argc, argv, shortopts, longopts, &indexptr );

      switch(ret)
	{
	case 0:
	  // long option that sets a flag
	case -1:
	  break;

	case 'h':
	  // help
	  print_help_message();
	  exit(0);
	  break;

	case 'n':
	  // number of first-stage tests
	  // this overrides the value from the header file
	  num_initial_tests = atoi(optarg);
	  break;

	case 'r':
	  // number of trials
	  // this overrides the value from the header file
	  num_repetitions = atoi(optarg);
	  break;

	case 'a':
	  // number of additional (second-stage) tests
	  // this overrides the value from the header file
	  num_additional_tests = atoi(optarg);
	  break;

	case 't':
	  // time limit for each trial
	  interface.set_max_time(atoi(optarg));
	  break;

	case 'o':
	  // output file
	  // Not yet implemented
	  cerr << "Option -o not yet implemented." << endl;
	  exit(1);
	  break;

	case 'c':
	  // recover from file
	  // Not yet implemented
	  cerr << "Option -c not yet implemented." << endl;
	  exit(1);
	  break;

	case 's':
	  // random seed
	  randgen.seed(atoi(optarg));
	  break;

	case 'e':
	  // Do exhaustive tests and write them to a file
	  exh_output.open(optarg);
	  use_exhaustive = true;
	  if (!exh_output.is_open())
	    {
	      cerr << "Failed to open file " << optarg << endl;
	      exit(1);
	    }
	  break;

	case 'i':
	  // read complete enumeration data from a file
	  exh_input.open(optarg);
	  use_from_exhaustive_file = true;
	  if (!exh_input.is_open())
	    {
	      cerr << "Failed to open file " << optarg << endl;
	      exit(1);
	    }
	  break;

	default:
	  cout << "Unknown option" << endl;
	  print_help_message();
	  exit(1);
	  break;
	}

    }
  while (ret != -1);

  // Postprocess integer variables that were set
  use_regression_tree = (regression_tree_int == 1);
  use_neural_net = ( neural_net_int== 1);
  use_naive = ( naive_int == 1);
  use_random = ( random_int == 1);
  use_our_heuristic = ( our_heuristic_int == 1);
  use_pairwise = ( pairwise_int == 1);
  use_double_pairwise = ( double_pairwise_int == 1);
  use_triplewise = ( triplewise_int == 1);

  // Check that options are consistent
  /// *** Need to finish this!

  // Defaults
  if (!use_random && !use_our_heuristic && !use_pairwise && 
      !use_double_pairwise && !use_triplewise && !use_exhaustive)
    use_our_heuristic = true;

  if (!use_regression_tree && !use_neural_net && !use_naive)
    use_naive = true;

  // Exhaustive tests
  if (use_exhaustive)
    {
      if (use_from_exhaustive_file)
	{
	  cerr << "Can''t write to and read from exhaustive file" << endl;
	  exit(1);
	}

      // Compute the total number of instances
      num_initial_tests = 1;
      for (int i = 0; i < num_parameters; i++)
	num_initial_tests *= par_levels[i];
      num_additional_tests = 0;
      cout << "PERFORMING EXHAUSTIVE TEST of " << num_initial_tests << " settings"
	   << endl;
    }

  // Consistency for naive
  if (use_naive)
    {
      // No additional tests
      num_additional_tests = 0;
    }

  // If neural net is wanted, FANN must be available
#ifdef NO_FANN
  if (use_neural_net)
    {
      cout << "You have requested a neural network, but STOP was compiled"
	   << endl << "with the NO_FANN flag." << endl;
      exit(1);
    }
#endif

  // If regression tree is wanted, check that R is available:
  if (use_regression_tree || (final_advice == 1) )
    {
      cout << "You have requested a regression-tree; checking whether R is available" << endl;
      if (system("R --version") != 0)
	{
	  cout << "R is not available; you cannot use regression-tree features" << endl;
	  exit(1);
	}
    }


  ////////////////////////////////////////////////////////////////
  // Option parsing is now done.  
  // After all the options, a single filename with filenames of instances
  //  to test
  if ((argc-optind) != 1)
    {
      print_help_message();
      exit(1);
    }
  // Read in list of test instances from file
  char * instance_filename = argv[optind];
  ifstream instance_file;
  instance_file.open(instance_filename);
  if (!instance_file.is_open())
    {
      cerr << "Failed to open file " << instance_filename << endl;
      exit(1);
    }
  vector<string> test_instances;
  while (!instance_file.eof() && !instance_file.fail())
    {
      // Read a filename
      string instance_name;
      instance_file >> instance_name;
      if (!instance_file.fail())
	test_instances.push_back(instance_name);
    }
  interface.set_test_instances( test_instances );

  // Write plan of experiments
  cout << "STOP experiment" << endl;
  cout << "Instances to test:" << endl;
  for (int i = 0; i < interface.get_test_instances().size(); i++)
    cout << "  " << interface.get_test_instances()[i] << endl;
  cout << "Initial tests: " << num_initial_tests << endl;
  cout << "Initial tests determined ";
  if (use_random)
    cout << "randomly" << endl;
  else if (use_our_heuristic)
    cout << "using conflict heuristic" << endl;
  else if (use_pairwise)
    cout << "using pairwise" << endl;
  else if (use_double_pairwise)
    cout << "using double-pairwise" << endl;
  else
    cout << "(not sure; omitted code)" << endl;

  if (use_naive)
    cout << "No machine learning to be done." << endl;
  else if (use_neural_net)
    cout << "A neural network will be trained on this data." << endl;
  else if (use_regression_tree)
    {
      cout << "A regression tree will direct additional tests." << endl;
      cout << "Additional tests: " << num_additional_tests << endl;
      cout << "Additonal tests determined ";
      if (use_random)
	cout << "randomly" << endl;
      else if (use_our_heuristic)
	cout << "using conflict heuristic" << endl;
      else
	cout << "using pairwise" << endl;
    }

  if (use_from_exhaustive_file)
    {
      cout << "Results are read from an exhaustive file rather than processed."<< endl;
    }

  // Read exhaustive file data (if using it)
  vector<double> exhaustive_input;
  if (use_from_exhaustive_file)
    {
      // Read in exhaustive file
      int npars;
      int temp;
      exh_input >> npars;
      if (npars != num_parameters)
	exit(1);
      for (int j = 0; j < num_parameters; j++)
	{
	  exh_input >> temp;
	  if (temp != par_levels[j])
	    exit(1);
	}
      // Compute the total number of settings
      int num_settings = 1;
      for (int i = 0; i < num_parameters; i++)
	num_settings *= par_levels[i];

      // Read in the instances
      // ** Assuming that they are in lexicographic order!
      double val;
      for (int i = 0; i < num_settings; i++)
	{
	  for (int j = 0; j < num_parameters; j++)
	    exh_input >> temp;
	  exh_input >> val;
	  exhaustive_input.push_back(val);
	}
      exh_input.close();  // done reading
      if (exhaustive_input.size() != num_settings)
	{
	  cerr << "Number of read instances incorrect" << endl;
	  exit(1);
	}
    }

  // Data from the initial test settings
  vector< vector<int> > initial_settings;
  vector< vector<int> > test_settings;
  vector< vector<int> > test_more_settings;
  int num_initial_settings;
  vector<double> objective_values (num_initial_tests);

  // Data from additional test settings
  vector< vector<int> > more_settings;
  int num_more_settings;
  vector<double> more_objective_values (num_additional_tests);


  ////////////////////////////////////////////////////////
  // Do the initial runs, or read the results from a file
  ////////////////////////////////////////////////////////

  cout << "Testing default setting" << endl;
  
  double scale_factor;
  double avg_first_stage;
  double avg_second_stage;
  double avg_overall;
  double total_first= 0;
  double total_second= 0;
  double total_overall= 0;

  double default_measure;
  if (use_from_exhaustive_file)
     default_measure = exhaustive_input[ exhaustive_index( default_settings, num_parameters, par_levels) ];
  else 
     default_measure = run_test_instances ( interface, default_settings );
  cout << " Measure for default setting: " << default_measure << endl;


  for (int m= 0; m < num_repetitions; m++)
  {
    double best_measure = BIG_DOUBLE;
    int best_setting = -1;
    double max_measure = 0;
    if (from_file == false)
    //////////////////////////////////////
    // Create initial settings and do initial runs
    ///////////////////////////////////////
    {
     // One trial is determined by a vector of integers.  

	if (use_exhaustive)
	{
	  initial_settings = exhaustive_settings( num_initial_tests, 
							num_parameters, par_levels);
          // Write the header to the output file
	  exh_output << num_parameters << endl;
	  for (int j = 0; j < num_parameters; j++)
		exh_output << par_levels[j] << " ";
	  exh_output << endl;
	}
	else if (use_our_heuristic)
	  initial_settings = limit_conflicts_settings( test_settings, num_initial_tests, 
								num_parameters,
								par_levels, randgen );
	else if (use_random)
	  initial_settings = random_settings( num_initial_tests, 
								num_parameters,
								par_levels, randgen );
	else if (use_pairwise)
	  initial_settings = pairwise_settings( num_initial_tests, num_parameters,
							par_levels, randgen, 2 );
	  
	

	num_initial_settings = initial_settings.size();
	  //  assert( num_initial_settings == num_initial_tests );

	// Loop through parameter settings
	for (int i = 0; i < num_initial_settings; i++ )
	{
	  double measure;
		  
	  if (use_from_exhaustive_file)
	    measure = exhaustive_input[ exhaustive_index( initial_settings[i], num_parameters, par_levels) ];
	  else  // this is the usual case
	    measure = run_test_instances ( interface, initial_settings[i] );

	// measure now has the statistic we care about.
	  objective_values[i] = measure;
	  cout << "Measure for these settings: " << measure << endl;

	// If doing exhaustive tests, write this to the file
	  if (use_exhaustive)
	  {
		// Output the settings
	    for (int j = 0; j < num_parameters; j++)
		exh_output << initial_settings[i][j] << " ";
			// Output the measure
	   exh_output << measure << endl;
	   }

	// Keep track of the best settings we see.
	  if (measure < best_measure)
	  {
	    best_measure = measure;
	    best_setting = i; // change this to indicate settings
	  }

	// Keep track of max measure
	  if (measure > max_measure)
	    max_measure = measure;

	} // main for-loop
	total_first += best_measure;
	if (use_exhaustive)
	  exh_output.close();  // done writing

    }

	#if 0  // This is for reading a FANN file
	else  
		/////////////////////////////////
		// Data is read from a file
		/////////////////////////////////
		{
		// Get the scale factor from the file.
		ifstream datafile(filename);
		int num_train;
		int num_in;
		int num_out;
		datafile >> num_train >> num_in >> num_out;
		if (num_in != num_inputs || num_out != 1)
		{
		cout << "Parameters for number of inputs and outputs do not match."
			<< endl << "Aborting." << endl;
		exit(1);
		}
		num_initial_tests = num_train;

		// Read in the training settings
		for (int j = 0; j < num_initial_tests; j++)
		{
		vector<int> settings;
		int temp;
		for (int i = 0; i < num_in; i++)
			{
			datafile >> temp;
			settings.push_back(temp);
			}
		vector<int> parameters = convert_input_to_params(settings, num_parameters, par_levels);
		initial_settings.push_back(parameters);

		datafile >> objective_values[j];

		// Keep track of the best settings from training
		if (objective_values[j] < best_measure)
			{
			best_measure = objective_values[j];
			best_setting = j;
			}
		}
		num_initial_settings = num_initial_tests;

		// Skip the remaining three newlines before the scale_factor
		for (int j = 0; j < 3; j++)
		datafile.ignore( 1024, '\n');
		datafile >> scale_factor;
		for (int j = 0; j < num_initial_tests; j++)
		objective_values[j] = objective_values[j] * scale_factor;
		best_measure = best_measure * scale_factor;

		datafile.close();
		}
	#endif  // reading from a FANN file


	/////////////////////////////////////  
	// The initial runs are now complete.  
	/////////////////////////////////////


	//////////////////////////////////////////////////
	// Do some "learning" and create additional tests
	//////////////////////////////////////////////////

	double best_prediction = BIG_DOUBLE;
	vector<int> predicted_best (num_parameters); 

#ifndef NO_FANN
  	if (use_neural_net && !use_regression_tree)
////////////////////////////////////////////////////////
// Neural network
////////////////////////////////////////////////////////
  	{

  // First determine the number of input nodes for the neural net.
	  num_inputs = 0;
	  input_starts.resize( num_parameters );
	  for (int i = 0; i < num_parameters; i++)
	  {
		input_starts[i] = num_inputs;
	  	if (par_levels[i] == 2)
		  num_inputs++;
	  	else if (par_levels[i] > 2)
		  num_inputs += par_levels[i];
	  }

	// Create the storage for input values
	  vector<vector<double> > input_values(num_initial_tests); 
	  for (int i = 0; i < num_initial_tests; i++) 
	  	input_values[i].resize(num_inputs);

	// Instantiate the neural network
	  const unsigned int num_layers = 3;
	  const unsigned int num_hidden = num_inputs;
	  const unsigned int num_output = 1;
	  struct fann * nnet;
	  if (use_neural_net)
	  {
	  	nnet = fann_create_standard( num_layers, num_inputs, 
					     num_hidden, num_output ); 
		fann_print_parameters(nnet);
	  }

	  // Write to a file for training
	  // First need to convert from settings to 'inputs'
	  for (int i = 0; i < num_initial_tests; i++)
	  {
	 	 vector<int> temp = convert_params_to_input (initial_settings[i], num_parameters, par_levels);
	  for (int j = 0; j < num_inputs; j++)
		  input_values[i][j] = temp[j];
	  // Note: objective_values[i] already has the corresponding value
	  }

	  // Scale so all values less than 1
	  scale_factor = (2 * max_measure);

	  // Write training data to a file for fann to use
	  ofstream datafile(filename);
	  // ??? should add error checking 
	  // Put Fann training data
	  datafile << num_initial_settings << " " << num_inputs << " 1" << endl;
	  for (int i = 0; i < num_initial_settings; i++ )
	  {
	    for (int j = 0; j < num_inputs; j++)
		datafile << input_values[i][j] << " ";
	   datafile << endl;
	   datafile << objective_values[i] / scale_factor << endl;
	  }
	// Hopefully, Fann will ignore the following lines.
	  datafile << "# The value below is a scale factor." << endl 
		<< "# It represents the measure corresponding to 1.0"
		<< endl;
	  // The final line contains the scale factor
	  datafile << scale_factor << endl;
	  datafile.close();

		
	  // The data is now ready for FANN to use
  
	  //  nnet.backprop_batch( input_values, objective_values);    
	  const int max_epochs = 5000;
	  const int epochs_between_reports = 1000;
	  const double desired_error = 0.000001 / scale_factor;
	  fann_train_on_file(nnet, filename, max_epochs, 
			     epochs_between_reports, desired_error); 

	  // Have the neural network predict the measure for all settings.
	  // To do this, need to loop through all settings.
	  // Store the results in a priority queue.
	  bool done = false;
	  vector<int> settings (num_parameters, 0);
	  priority_queue< pair< double, vector<int> > > measures;

	  // ??? Temporary kludge:
	  fann_type * input_copy = new fann_type[ num_inputs ];

	  while (done == false)
	  {
       	 	 vector<int> input = convert_params_to_input(settings, 
							     num_parameters, 
							     par_levels);

		 // ??? kludge
	  	for (int i = 0; i < num_inputs; i++)	
		  input_copy[i] = input[i];

		// double value = nnet.evaluator( input );
       	  	fann_type *val;
	  	val = fann_run( nnet, input_copy );
		double value = *val;

		 // Put this result in the priority queue
		 // Use the opposite of the value, since priority queues 
		 // prefer higher values.
		 measures.push( pair< double, vector<int> > (-value, settings) );

#if 0  // this is for testing the neural net's prediction for all values
		// Compute true best
		double actual_measure = 
			run_test_instances ( interface, settings );
#endif

#if 0  // this is for recording the best prediction
	 	 if (value < best_prediction)
	         {
		   best_prediction = value;
		   for (int j = 0; j < num_parameters; j++)
		  	predicted_best[j] = settings[j];
	 	 }
	  	for (int k = 0; k < num_parameters; k++)
			cout << settings[k] << " ";
	  	cout << ": " << value * scale_factor << endl;
			//	       << "  " << actual_measure << endl;
#endif	      
	      
		// increment settings
	  	int index = num_parameters - 1;
	  	while ( (index >= 0) && (settings[index] == par_levels[index] - 1) )  
			index--;
	  	if (index == -1) 
			done = true;
	  	else
    	  	{
	  	  settings[index]++;
	  	  index++;
   	  	  while (index < num_parameters)
	  	  	settings[index++] = 0;
	 	}
  	  }

	  // Create a vector of vectors containing the top selections
	  vector<vector<int> > choices;

	  // For now, keep top 10%
	  int num_candidates = measures.size() / 10;
	  for (int i = 0; i < num_candidates; i++)
	    {
	      pair<double, vector<int> > temp = measures.top();
	      vector<int> temp_vec = temp.second;
	      choices.push_back( temp_vec );
	      measures.pop();
	      // For debugging
	      // cout << " candidate estimate: " << -temp.first << endl;
	    }

	  // Use our conflicts heuristic to select a subset of these
	  // for more_settings
	  more_settings = limit_conflicts_settings_subset( num_additional_tests, 
							   num_parameters,
							   par_levels, 
							   randgen, choices );

#if 0   // This is another evaluation of the neural net performance
	// Compare prediction with the best we saw
	measure = 
	run_test_instances ( interface, predicted_best );

	// Output results

	// Report on how well the training instances are now "predicted"
	cout << "Report on training instances:" << endl;
	for (int i = 0; i < num_initial_settings; i++ )
	{
		vector<int> input = convert_params_to_input(initial_settings[i], 
								num_parameters, 
								par_levels);
	// ??? kludge
	for (int j = 0; j < num_inputs; j++)
		input_copy[j] = input[j];

	// double value = nnet.evaluator( input );
	fann_type *val;
	val = fann_run( nnet, input_copy );
	double value = (*val) * scale_factor;

	for (int k = 0; k < num_parameters; k++)
		cout << initial_settings[i][k] << " ";
	cout << ": predicted " << value 
			<< ", actual " << objective_values[i] << endl;

		}  
#endif // checking the best prediction

	fann_destroy(nnet);

  	} // end of neural net loop
#endif

	if (use_regression_tree && !use_neural_net)
		///////////////////////////////////////////
		// Regression tree
		///////////////////////////////////////////
	{
		//// Write a data file for R
		ofstream datafile("R_data");

		// Header
		for (int i = 0; i < num_parameters; i++)
			datafile << "P" << i << " ";
		datafile << "Measure" << endl;

		// Data
		for (int i = 0; i < num_initial_settings; i++ )
		{
			for (int j = 0; j < num_parameters; j++)
			datafile << initial_settings[i][j] << " ";
			datafile << objective_values[i] << endl;
		}

		datafile.close();

		//// Create R script file
		ofstream scriptfile("R_script");

		scriptfile << "inp <- read.table (\"R_data\", header = TRUE)" << endl;
		for (int i = 0; i < num_parameters; i++)
			scriptfile << "inp$P" << i << " <- factor(inp$P" << i
			<< ", levels = c(0:" << par_levels[i] 
			<< "), ordered = FALSE)" << endl;

		scriptfile << "library(rpart)" << endl;
		scriptfile << "treemodel <- rpart (Measure ~ P0";
		for (int i = 1; i < num_parameters; i++)
			scriptfile << " + P" << i;
		scriptfile << ", data = inp, method = \"anova\", " 
			<< "control = rpart.control(minsplit = 3, maxcompete = 0,"
			<< " maxsurrogate= 0, cp=0) )" << endl;
		// The next part depends on how many branchings are being 
		// requested.
		// We want the nodes that are powers of 2, starting with 2.
		int node = 1;
		for (int i = 1; i <= num_regression_tree_branchings; i++)
		  {
		    node *= 2;  // should be 2^i
		    scriptfile << "if (!is.na(treemodel$frame[\"" 
			       << node << "\",1])) {" << endl;
		    scriptfile << "  cat(path.rpart(treemodel, " << node 
			       << ")$\"" << node << "\"[[" << i+1 
			       << "]],\"\n\", file=\"R_out\",append=";
		    if (i==1) scriptfile << "FALSE";
		    else      scriptfile << "TRUE";
		    scriptfile << ")" << endl << "}" << endl;
		  }
		scriptfile.close();

		//// Run R externally
		if (system("R --vanilla < R_script") != 0)
		{
			cerr << "Error returned when attempting system call to run R" << endl;
			exit(1);
		}
		// output should have been placed in "R_out"

		//// Extract the result from R
		// Format is "P%d=%d,%d,%d..."
		ifstream resultfile("R_out");
		char buffer[128];
		char * token;
		vector<int> branch_var;
		int temp;
		vector<vector<int> > branch_val;
		char delims[] = "P=,";

		while (!resultfile.eof())
		  {
		    resultfile.getline( buffer, 128 );
		    if (resultfile.eof()) break;

		    branch_var.push_back( atoi(strtok(buffer, delims)) );
		    token = strtok(NULL, delims);
		    branch_val.resize(branch_var.size());
		    do
		      {
			temp = atoi(token);
			branch_val[branch_val.size()-1].push_back(temp);
			token = strtok(NULL, delims);
		      }
		    while (token != NULL);
		  }

		resultfile.close();

		cout << "Number of variables found: " << branch_var.size()
		     << endl;
		for (int i = 0; i < branch_var.size(); i++)
		  {
		    cout << "Var " << i+1 << ": " << branch_var[i] 
			 << " with values";
		    for (int j = 0; j < branch_val[i].size(); j++)
		      cout << " " << branch_val[i][j];
		    cout << endl;
		  }


		//// We now have the values we need.
		//// Decide on additional tests to perform.

		// These tests uses what the regression tree learned, but ignore
		// all the initial tests when designing the new tests.

		// Modify the par_levels accordingly
		vector<int> mod_par_levels = par_levels;
		for (int i = 0; i < branch_var.size(); i++)
		  mod_par_levels[branch_var[i]] = branch_val[i].size();

		if (use_our_heuristic)
			more_settings = limit_conflicts_settings( test_more_settings, num_additional_tests, 
								num_parameters,
								mod_par_levels, 
								randgen );
		else if (use_random)
			more_settings = random_settings( num_additional_tests, 
							num_parameters,
							mod_par_levels, 
							randgen );
		/*else if (use_pairwise)
			more_settings = pairwise_settings( num_additional_tests, num_parameters,
							mod_par_levels, randgen, 2 );*/
		else if (use_pairwise)
		  more_settings = pairwise_settings( num_additional_tests, num_parameters,
							mod_par_levels, randgen, 2 );


		//cout << "-----------Corrected settings:" << endl;
		// Now we have to correct settings for var1 and var2.
		for (int i=0; i < more_settings.size(); i++)
		{
		  for (int j=0; j < branch_val.size(); j++)
		    {
			more_settings[i][branch_var[j]] = 
			  branch_val[j][more_settings[i][branch_var[j]]];
		    }
			
		}

	} // end regression tree

	if (use_regression_tree && use_neural_net)
	{
		cout << "Using regression tree on neural net predictions not yet implemented." << endl;
	}


	/////////////////////////////////////////
	// Do additional tests based on learning.
	/////////////////////////////////////////

	// For now, data is not stored.

	double best_more_measure = BIG_DOUBLE;
	int best_more_setting = -1;

	num_more_settings = more_settings.size();
	//  assert( num_more_settings == num_additional_tests );

	// Loop through parameter settings
	for (int i = 0; i < num_more_settings; i++ )
	{
		double measure;
		if (use_from_exhaustive_file)
			measure = exhaustive_input[ exhaustive_index( more_settings[i], num_parameters, par_levels) ];
		else  // this is the usual case
			measure = run_test_instances ( interface, more_settings[i] );

		// measure now has the statistic we care about.
		cout << "Measure for these settings: " << measure << endl;

		// Keep track of the best settings we see.
		if (measure < best_more_measure)
		{
			best_more_measure = measure;
			best_more_setting = i; // change this to indicate settings
		}

	} // main for-loop
	total_second += best_more_measure;
	if (best_more_measure > best_measure)
		total_overall += best_measure;
	else
		total_overall += best_more_measure;

	///////////////////////////////////////////
	// Done with all tests.  Output findings.
	///////////////////////////////////////////

	// Output best of parameters 
	cout << "Best measure from initial tests is " << best_measure << endl;
	cout << "Best settings from initial tests were " << endl;  

	vector<int> best_par = initial_settings[best_setting];
	for (int k = 0; k < num_parameters; k++)
	{
		cout << "  " << par_name[k] << ": " 
		<< best_par[k] << ", " << par_level[k][best_par[k]] << endl;
	}

	if (num_additional_tests > 0)
	{
		cout << "Best measure from second tests is " << best_more_measure << endl;
		cout << "Best settings from second tests were " << endl;  

		best_par = more_settings[best_more_setting];
		for (int k = 0; k < num_parameters; k++)
		{
		cout << "  " << par_name[k] << ": " 
			<< best_par[k] << ", " << par_level[k][best_par[k]] << endl;
		}
	}

	if (final_advice == 1)
	  ///////////////////////////////////////////
	  // Regression tree for final advice
	  ///////////////////////////////////////////
	  {
	    // For this, we might as well report lots of advice
	    num_regression_tree_branchings = 5;

	    //// Write a data file for R
	    ofstream datafile("R_data");

	    // Header
	    for (int i = 0; i < num_parameters; i++)
	      datafile << "P" << i << " ";
	    datafile << "Measure" << endl;

	    // Data
	    for (int i = 0; i < num_initial_settings; i++ )
	      {
		for (int j = 0; j < num_parameters; j++)
		  datafile << initial_settings[i][j] << " ";
		datafile << objective_values[i] << endl;
	      }
	    for (int i = 0; i < num_more_settings; i++ )
	      {
		for (int j = 0; j < num_parameters; j++)
		  datafile << more_settings[i][j] << " ";
		datafile << more_objective_values[i] << endl;
	      }

	    datafile.close();

	    //// Create R script file
	    ofstream scriptfile("R_script");

	    scriptfile << "inp <- read.table (\"R_data\", header = TRUE)" << endl;
	    for (int i = 0; i < num_parameters; i++)
	      scriptfile << "inp$P" << i << " <- factor(inp$P" << i
			 << ", levels = c(0:" << par_levels[i] 
			 << "), ordered = FALSE)" << endl;

	    scriptfile << "library(rpart)" << endl;
	    scriptfile << "treemodel <- rpart (Measure ~ P0";
	    for (int i = 1; i < num_parameters; i++)
	      scriptfile << " + P" << i;
	    scriptfile << ", data = inp, method = \"anova\", " 
		       << "control = rpart.control(minsplit = 3, maxcompete = 0,"
		       << " maxsurrogate= 0, cp=0) )" << endl;
	    // The next part depends on how many branchings are being 
	    // requested.
	    // We want the nodes that are powers of 2, starting with 2.
	    int node = 1;
	    for (int i = 1; i <= num_regression_tree_branchings; i++)
	      {
		node *= 2;  // should be 2^i
		scriptfile << "if (!is.na(treemodel$frame[\"" 
			   << node << "\",1])) {" << endl;
		scriptfile << "  cat(path.rpart(treemodel, " << node 
			   << ")$\"" << node << "\"[[" << i+1 
			   << "]],\"\n\", file=\"R_out\",append=";
		if (i==1) scriptfile << "FALSE";
		else      scriptfile << "TRUE";
		scriptfile << ")" << endl << "}" << endl;
	      }
	    scriptfile.close();

	    //// Run R externally
	    if (system("R --vanilla < R_script") != 0)
	      {
		cerr << "Error returned when attempting system call to run R" << endl;
		exit(1);
	      }
	    // output should have been placed in "R_out"

	    //// Extract the result from R
	    // Format is "P%d=%d,%d,%d..."
	    ifstream resultfile("R_out");
	    char buffer[128];
	    char * token;
	    vector<int> branch_var;
	    int temp;
	    vector<vector<int> > branch_val;
	    char delims[] = "P=,";

	    while (!resultfile.eof())
	      {
		resultfile.getline( buffer, 128 );
		if (resultfile.eof()) break;
		
		branch_var.push_back( atoi(strtok(buffer, delims)) );
		token = strtok(NULL, delims);
		branch_val.resize(branch_var.size());
		do
		  {
		    temp = atoi(token);
		    branch_val[branch_val.size()-1].push_back(temp);
		    token = strtok(NULL, delims);
		  }
		while (token != NULL);
	      }

	    resultfile.close();

	    cout << "Final advice from regression tree." << endl;
	    cout << "Up to " << num_regression_tree_branchings 
		 << " branchings will be reported." << endl;
	    cout << "  Number of branchings in tree: " << branch_var.size()
		 << endl;
	    for (int i = 0; i < branch_var.size(); i++)
	      {
		cout << "Var " << i+1 << ": " << branch_var[i] 
		     << " with values";
		for (int j = 0; j < branch_val[i].size(); j++)
		  cout << " " << branch_val[i][j];
		cout << endl;
	      }

	  }


  }
  if (num_additional_tests > 0)
    {
      avg_first_stage = total_first/  num_repetitions;
      cout << "Best measure for the first stage tests is "<< avg_first_stage << endl;
      avg_second_stage = total_second/  num_repetitions;
      cout << "Best measure for the second stage tests is "<< avg_second_stage << endl;
    }
  avg_overall = total_overall/  num_repetitions;
  cout << "Best measure for all the tests is "<< avg_overall << endl;

  cout << "Measure for default settings is "<< default_measure << endl;
/*
  cout << "Best settings from all tests were " << endl;  

		best_par = more_settings[best_more_setting];
		for (int k = 0; k < num_parameters; k++)
		{
		cout << "  " << par_name[k] << ": " 
			<< best_par[k] << ", " << par_level[k][best_par[k]] << endl;
		}	
*/

}


double 
run_test_instances ( STOP_interface & interface, vector<int> setting )
{

  int num_instances = interface.get_num_instances();
  vector<string> test_instances = interface.get_test_instances();

  // Set the given parameters
  // Parameter settings are defined by a vector of integers
  int ret = interface.set_parameters( setting );

  // Initialize the performance metric
  interface.initialize_metric();

  // Loop through test instances
  for (int j = 0; j < num_instances ; j++ )
    {
      // Run the instance
      ret = interface.load_instance(test_instances[j]);
      ret = interface.execute_program();
	  
      if (ret != 0) // something wrong during execution
	{
	  cerr << "ERROR: Return value " << ret 
	       << " when executing solver on instance " 
	       << test_instances[j] << endl;
	}
    } // instance for-loop

  // Return the metric
  return interface.get_metric();
}

vector<int> 
convert_params_to_input (const vector<int> & settings, 
			 const int num_parameters, 
			 const vector<int> & par_levels)
{
  vector<int> nn_input (num_inputs, 0);

  for (int j = 0; j < num_parameters; j++)
    {
      if (par_levels[j] == 2)
	nn_input[input_starts[j]] = settings[j]; // should be 0 or 1
      else if (par_levels[j] > 2)
	nn_input[input_starts[j] + settings[j] ] = 1;  
        // others will remain at 0
    }

  return nn_input;
}

vector<int>
convert_input_to_params (const vector<int> & nn_input, 
			 const int num_parameters,
			 const vector<int> & par_levels)
{
  vector<int> settings (num_parameters, 0);

  for (int j = 0; j < num_parameters; j++)
    {
      if (par_levels[j] == 2)
	settings[j] = nn_input[input_starts[j]]; // should be 0 or 1
      else if (par_levels[j] > 2)
	{
	  // find the index with a 1
	  int index = input_starts[j];
	  while (nn_input[index] == 0) index++;

	  settings[j] = index - input_starts[j];
	  // should check that settings[j] < par_levels[j]
	}
    }

  return settings;
}


#ifndef WIN32
double 
time_now ( void )
{
  struct rusage usage;
  struct rusage child_usage;
  int ret = getrusage(RUSAGE_SELF, &usage);
  int ret_child;
  if (ret == 0)
    ret_child = getrusage(RUSAGE_CHILDREN, &child_usage);
  if (ret != 0 || ret_child != 0)
    {
      cout << "Error with measuring time.  Aborting." << endl;
      exit(1);
    }
  double utime = usage.ru_utime.tv_sec + 1e-6 * usage.ru_utime.tv_usec;
  double child_utime = child_usage.ru_utime.tv_sec + 1e-6 * child_usage.ru_utime.tv_usec;

  return (utime + child_utime);
}
#endif

#ifdef WIN32
double
time_now (void)
{
  //***** In Windows, apparently only wall-clock time is possible.
	double value = clock();
	//	printf(" Checking time: %f / %ld = ", value, CLOCKS_PER_SEC);
	value /= CLOCKS_PER_SEC;
	//	printf("%f\n", value);
	return value;
}
#endif

// The following function takes a setting as a vector<int> and returns
// a single index for the setting, corresponding to that setting's
// position in an exhaustive lexicographic order.
int 
exhaustive_index ( vector<int> setting,
		   const int num_parameters, const vector<int> & par_levels )
{
  int index = 0;
  for (int j = num_parameters-1; j >= 0; j--)
    {
      int multiplier = 1;
      for (int k = num_parameters-1; k > j; k--)
	multiplier *= par_levels[k];
      index += multiplier * setting[j];
    }
  return index;
}

// Display information about options
void 
print_help_message ( void )
{
  cout << "Usage: STOP [options] <file_of_instances>" << endl << endl;
  cout << "Available options:" << endl << endl;
  cout << "  Options for design of initial trials" << endl;
  cout << "   --heuristic   (conflicts heuristic) [default]" << endl;
  cout << "   --random" << endl;
  cout << "   --pairwise" << endl;

  cout << endl << "  Options for machine learning" << endl;
  cout << "   --naive  (no learning; no additional tests) [default]" << endl;
  cout << "   --regression-tree" << endl;
  cout << "   --neural-net" << endl;

  cout << endl << "  Options for number of tests" << endl;
  cout << "   -n #  --number-initial-trials #  [default is in interface file]" << endl;
  cout << "   -a #  --number-additional-trials #  [default is in interface file]" << endl;

  cout << endl << "  Other options" << endl;
  cout << "   -s #  --random-seed #  [default is a random random seed]" << endl;
  cout << "   --advice  (use regression tree at end to provide advice)" << endl;

  cout << endl << "  Development options; not recommended unless you know what you're doing" << endl;
  cout << "   --input-file <filename>  (file previously created with --exhaustive)" << endl;
  cout << "   --exhaustive <filename>  (do exhaustive tests and write to file)" << endl;
  cout << "   -r #  --number-replicates #  [default is in interface file]" << endl;

}
