//  interface_cbc.hpp
//
//  Copyright 2007 University of Pittsburgh.
//  U of Pittsburgh coding done by Brady Hunsaker and Mustafa Baz.
//
//  This file is part of STOP (Selection Tool for Optimization Parameters).
//
//  The contents of this file are subject to the Common Public License
//  1.0.  (the "License"); you may not use this file except in
//  compliance with the License. You should have received a copy of
//  the Common Public License along with STOP.
//
//  Software distributed under the License is distributed on an "AS
//  IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
//  implied. See the License for the specific language governing
//  rights and limitations under the License.
//
//  Alternatively, the contents of this file may be used under the
//  terms of the GNU General Public License Version 2 or later (the
//  "GPL"), in which case the provisions of the GPL are applicable
//  instead of those above. If you wish to allow use of your version
//  of this file only under the terms of the GPL, and not to allow
//  others to use your version of this file under the terms of the
//  CPL, indicate your decision by deleting the provisions above and
//  replace them with the notice and other provisions required by the
//  GPL. If you do not delete the provisions above, a recipient may
//  use your version of this file under the terms of either the CPL or
//  the GPL.
//
/////////////////////////////////////////////////////////////

// Additional include files go here
#include <stdio.h>
#include <vector>
#include <string>
#include <cassert>

using namespace std;

#define BIG_DOUBLE 1.0e10;

// function in STOP for timing
double time_now( void );


class STOP_interface
{

public:

  // Initialization is all done in the default constructor.
  STOP_interface ();
  ~STOP_interface ();

  // Metric for testing
  void initialize_metric( void );
  double get_metric( void );

  //  Function to set the parameters
  int set_parameters ( vector<int> setting );
  //  Function to load an instance
  int load_instance ( string filename );
  //  Function to call the program
  int execute_program ( void );

  // Find out the values of constants
  int get_num_initial_tests( void );
  int get_num_repetitions( void );
  int get_num_additional_tests( void );
  int get_num_parameters( void );
  vector<string> get_par_name( void );
  vector<int> get_par_levels( void );
  vector<int> get_default_settings( void );
  vector< vector<string> > get_par_level( void );
  
  // Max time for each trial
  void set_max_time( int );
  int get_max_time( void );

  // Test instances
  void set_test_instances( vector<string> );
  vector<string> get_test_instances( void );
  int get_num_instances( void );

private:
  ////////////////////////////////////////////////////
  // Any additional variables that are necessary should be defined here.

  // We need a pointer to the instance
  vector<int> current_setting;
  string command;

  ////////////////////////////////////////////////////
  // Standard variables.  These are all set in the constructor.
  int num_initial_tests;
  int num_repetitions;
  int num_additional_tests;
  int max_time;
  char time_str[128];
  int num_parameters;
  vector<string> par_name;
  vector<int> par_levels;
  vector<int> default_settings;
  vector< vector<string> > par_level;
  vector<string> test_instances;

  // The metric for comparing settings
  double metric;

}; // end of class



////////////////////////////////////////////////////
// Function definitions

// The default constructor does two things:
//   - sets constant values
//   - does any initialization that may be required by the program

STOP_interface::STOP_interface ()
{
  // Set constant values
  // NOTE: these may be overridden at the command line
  num_initial_tests = 16;
  num_repetitions = 1;
  num_additional_tests = 16;
  max_time = 86400; // 24 hours
  sprintf(time_str,"%d", max_time);

  // Give a name to each parameter and store the number of levels each has
  par_name.push_back ( "STRONG_BRANCHING" );
  par_levels.push_back(3);
  default_settings.push_back(1);
  par_name.push_back ("COST_STRATEGY");
  par_levels.push_back(2);
  default_settings.push_back(0);
  par_name.push_back ("GOMORY_CUTS");
  par_levels.push_back(4);
  default_settings.push_back(3);
  par_name.push_back ("MIR_CUTS");
  par_levels.push_back(4);
  default_settings.push_back(3);
  par_name.push_back ("PROBING");
  par_levels.push_back(4);
  default_settings.push_back(3);
  par_name.push_back ("CLIQUE_CUTS");
  par_levels.push_back(4);
  default_settings.push_back(3);
  par_name.push_back ("FEASIBILITY");
  par_levels.push_back(2);
  default_settings.push_back(1);

  num_parameters = par_name.size();
  assert( par_levels.size() == num_parameters );

  // For each parameter, give a name to each setting 

  par_level.resize( num_parameters );

  par_level[0].push_back("0");
  par_level[0].push_back("5");
  par_level[0].push_back("9");
  assert( par_level[0].size() == par_levels[0] );

  par_level[1].push_back("off");
  par_level[1].push_back("priorities");
  assert( par_level[1].size() == par_levels[1] );

  par_level[2].push_back("off");
  par_level[2].push_back("on");
  par_level[2].push_back("root");
  par_level[2].push_back("ifmove");
  assert( par_level[2].size() == par_levels[2] );

  par_level[3].push_back("off");
  par_level[3].push_back("on");
  par_level[3].push_back("root");
  par_level[3].push_back("ifmove");
  assert( par_level[3].size() == par_levels[3] );

  par_level[4].push_back("off");
  par_level[4].push_back("on");
  par_level[4].push_back("root");
  par_level[4].push_back("ifmove");
  assert( par_level[4].size() == par_levels[4] );

  par_level[5].push_back("off");
  par_level[5].push_back("on");
  par_level[5].push_back("root");
  par_level[5].push_back("ifmove");
  assert( par_level[5].size() == par_levels[5] );


  par_level[6].push_back("off");
  par_level[6].push_back("on");
  assert( par_level[6].size() == par_levels[6] );

  /////////////////////////////////////
  // Now do any initialization required by the program.

  // Nothing necessary for CBC
}

// This function does any necessary initialization of the metric.
void 
STOP_interface::initialize_metric( void )
{
  // Our metric is total time.  Reset to zero before running the 
  // set of instances.
  metric = 0.0;
}

// Report the metric value to the main program
double 
STOP_interface::get_metric( void )
{
  return metric;
}


// This function sets the parameters based on a vector of integers.
int 
STOP_interface::set_parameters ( vector<int> setting )
{
  // Unfortunately, GLPK can't change these settings until the problem 
  // is loaded, so just copy the setting into a global variable and do it
  // later.
  current_setting = setting;

  return 0;
}


//  This function loads an instance based on a string containing 
// the name of the instance.
int 
STOP_interface::load_instance ( string filename )
{
  
// Set the string here
  command = "./cbc " + filename + 
    " -strong " + par_level[0][current_setting[0]] +
    " -cost " + par_level[1][current_setting[1]] + 
    " -gomory " + par_level[2][current_setting[2]] +
    " -mixed " + par_level[3][current_setting[3]] +
    " -probing " + par_level[4][current_setting[4]] + 
    " -clique " + par_level[5][current_setting[5]] +
    " -feas " + par_level[6][current_setting[6]] + 
    " -sec " + time_str +  
    " -solve";
  return 0;
}


//  This function runs the test program.
int 
STOP_interface::execute_program ( void )
{
  double time_start = time_now();
  int ret = system(command.c_str());
  double time_stop = time_now();
  double cur_time = time_stop - time_start;

  // *** We should really do some sort of check here to see whether the output 
  // really claims to be an optimal solution.

  metric += cur_time;

  return ret;
}


// Functions to return values of constants

int 
STOP_interface::get_num_initial_tests( void )
{
  return num_initial_tests;
}

int 
STOP_interface::get_num_repetitions( void )
{
  return num_repetitions;
}

int 
STOP_interface::get_num_additional_tests( void )
{
  return num_additional_tests;
}

int
STOP_interface::get_max_time( void )
{
  return max_time;
}

void
STOP_interface::set_max_time( int time )
{
  if (time > 0)  max_time = time;
}

int 
STOP_interface::get_num_parameters( void )
{
  return num_parameters;
}

vector<string> 
STOP_interface::get_par_name( void )
{
  return par_name;
}

vector<int> 
STOP_interface::get_par_levels( void )
{
  return par_levels;
}

vector<int> 
STOP_interface::get_default_settings( void )
{
  return default_settings;
}

vector< vector<string> > 
STOP_interface::get_par_level( void )
{
  return par_level;
}

int
STOP_interface::get_num_instances( void )
{
  return test_instances.size();
}

vector<string> 
STOP_interface::get_test_instances( void )
{
  return test_instances;
}

void
STOP_interface::set_test_instances( vector<string> instances)
{
  test_instances = instances;
}

// Destructor
STOP_interface::~STOP_interface( )
{
  // Nothing needs to be done
}
