//  interface_cplex.hpp
//
//  Copyright 2007 University of Pittsburgh.
//  U of Pittsburgh coding done by Brady Hunsaker and Mustafa Baz.
//
//  This file is part of STOP (Selection Tool for Optimization Parameters).
//
//  The contents of this file are subject to the Common Public License
//  1.0.  (the "License"); you may not use this file except in
//  compliance with the License. You should have received a copy of
//  the Common Public License along with STOP.
//
//  Software distributed under the License is distributed on an "AS
//  IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
//  implied. See the License for the specific language governing
//  rights and limitations under the License.
//
//  Alternatively, the contents of this file may be used under the
//  terms of the GNU General Public License Version 2 or later (the
//  "GPL"), in which case the provisions of the GPL are applicable
//  instead of those above. If you wish to allow use of your version
//  of this file only under the terms of the GPL, and not to allow
//  others to use your version of this file under the terms of the
//  CPL, indicate your decision by deleting the provisions above and
//  replace them with the notice and other provisions required by the
//  GPL. If you do not delete the provisions above, a recipient may
//  use your version of this file under the terms of either the CPL or
//  the GPL.
//
/////////////////////////////////////////////////////////////

// Additional include files go here
extern "C" {
#include "cplex.h"
}

#include <vector>
#include <string>
#include <cassert>
#include <iostream>

using namespace std;

#define BIG_DOUBLE 1.0e10;

// function in STOP for timing
double time_now( void );


class STOP_interface
{

public:

  // Initialization is all done in the default constructor.
  STOP_interface ();
  ~STOP_interface ();

  // Metric for testing
  void initialize_metric( void );
  double get_metric( void );

  //  Function to set the parameters
  int set_parameters ( vector<int> setting );
  //  Function to load an instance
  int load_instance ( string filename );
  //  Function to call the program
  int execute_program ( void );

  // Find out the values of constants
  int get_num_initial_tests( void );
  int get_num_repetitions( void );
  int get_num_additional_tests( void );
  int get_num_parameters( void );
  vector<string> get_par_name( void );
  vector<int> get_par_levels( void );
  vector<int> get_default_settings( void );
  vector< vector<string> > get_par_level( void );

  // Max time for each trial
  void set_max_time( int );
  int get_max_time( void );

  // Test instances
  void set_test_instances( vector<string> );
  vector<string> get_test_instances( void );
  int get_num_instances( void );


private:
  ////////////////////////////////////////////////////
  // Any additional variables that are necessary should be defined here.

  // We need a pointer to the instance and environment
  CPXENVptr env;
  CPXLPptr instance;

  ////////////////////////////////////////////////////
  // Standard variables.  These are all set in the constructor.
  int num_initial_tests;
  int num_repetitions;
  int num_additional_tests;
  int max_time;
  int num_parameters;
  vector<string> par_name;
  vector<int> par_levels;
  vector<int> default_settings;
  vector< vector<string> > par_level;

  vector<string> test_instances;

  // The metric for comparing settings
  double metric;

}; // end of class



////////////////////////////////////////////////////
// Function definitions

// The default constructor does two things:
//   - sets constant values
//   - does any initialization that may be required by the program

STOP_interface::STOP_interface ()
{
  // Set default constant values
  // NOTE: these may be overridden at the command line
  num_initial_tests = 16;
  num_repetitions = 1;
  num_additional_tests = 16;
  max_time = 86400; // 24 hours

  // Main parameters
  // CPX_PARAM_MIPEMPHASIS 0, 1, 2, 3, 4
  // CPX_PARAM_NODESEL 1, 2, 3, ( 0 )
  // CPX_PARAM_VARSEL  0, 2, 3 (-1, 1, 4)
  // CPX_PARAM_DIVETYPE 0, 1, 2, 3
  // CPX_PARAM_FRACCUTS 0, 1, -1, (2)
  // CPX_PARAM_MIRCUTS 0, 1, -1, (2)

  // additional parameters
  // CPX_PARAM_DISJCUTS  -1, 0, 1 
  // CPX_PARAM_CLIQUES  -1, 0, 1
  // CPX_PARAM_PRESLVND -1, 0, 1
  // CPX_PARAM_PROBE -1, 0, 1
  // CPX_PARAM_MIPORDTYPE 0, 1, 2, 3 
  // CPX_PARAM_BRDIR -1, 0, 1,

  // Give a name to each parameter and store the number of levels each has
  par_name.push_back ( "MIP emphasis" );
  par_levels.push_back(5);
  default_settings.push_back(0);
  par_name.push_back ("Node selection");
  par_levels.push_back(3);
  default_settings.push_back(0);
  par_name.push_back ("Branching variable");
  par_levels.push_back(3);
  default_settings.push_back(0);
  par_name.push_back ("Dive type");
  par_levels.push_back(4);
  default_settings.push_back(0);
  par_name.push_back ("Fractional cuts");
  par_levels.push_back(3);
  default_settings.push_back(1);
  par_name.push_back ("MIR cuts");
  par_levels.push_back(3);
  default_settings.push_back(1);

/*
  par_name.push_back ("Disjunctive cuts");
  par_levels.push_back(3);
  default_settings.push_back(1);
  par_name.push_back ("Clique cuts");
  par_levels.push_back(3);
  default_settings.push_back(1);
  par_name.push_back ("Node presolve selector");
  par_levels.push_back(3);
  default_settings.push_back(1);
  par_name.push_back ("Probing");
  par_levels.push_back(3);
  default_settings.push_back(1);
 
*/
/*  par_name.push_back ("MIP priority");
  par_levels.push_back(4);
  default_settings.push_back();
  par_name.push_back ("MIP branching direction");
  par_levels.push_back(3);
  default_settings.push_back();

  
*/
num_parameters = par_name.size();

  // For each parameter, give a name to each setting 
  par_level.resize( num_parameters );

  // MIP emphasis
  par_level[0].push_back("automatic");
  par_level[0].push_back("feasibility");
  par_level[0].push_back("optimality");
  par_level[0].push_back("moving best bound");
  par_level[0].push_back("hidden feasibility");
  assert( par_level[0].size() == par_levels[0] );

  // Node selection
  par_level[1].push_back("best-bound");
  par_level[1].push_back("best-estimate");
  par_level[1].push_back("alternative best-estimate");
  assert( par_level[1].size() == par_levels[1] );

  // Branching variable selection
  par_level[2].push_back("automatic");
  par_level[2].push_back("pseudo costs");
  par_level[2].push_back("strong branching");
  assert( par_level[2].size() == par_levels[2] );

  // Dive type
  par_level[3].push_back("automatic");
  par_level[3].push_back("traditional dive");
  par_level[3].push_back("probing dive");
  par_level[3].push_back("guided dive");
  assert( par_level[3].size() == par_levels[3] );

  // Fractional cuts
  par_level[4].push_back("off");
  par_level[4].push_back("automatic");
  par_level[4].push_back("on");
  assert( par_level[4].size() == par_levels[4] );

  // MIR cuts
  par_level[5].push_back("off");
  par_level[5].push_back("automatic");
  par_level[5].push_back("on");
  assert( par_level[5].size() == par_levels[5] );

/*
  // Disjunctive cuts
  par_level[6].push_back("off");
  par_level[6].push_back("automatic");
  par_level[6].push_back("on");

  assert( par_level[6].size() == par_levels[6] );

  // Clique cuts
  par_level[7].push_back("off");
  par_level[7].push_back("automatic");
  par_level[7].push_back("on");
  assert( par_level[7].size() == par_levels[7] );

 // Node presolve selector
  par_level[8].push_back("off");
  par_level[8].push_back("automatic");
  par_level[8].push_back("force");
  assert( par_level[8].size() == par_levels[8] );

  // MIP Probe
  par_level[9].push_back("off");
  par_level[9].push_back("automatic");
  par_level[9].push_back("on");
  assert( par_level[9].size() == par_levels[9] );

*/


/* // MIP priority
  par_level[8].push_back("off");
  par_level[8].push_back("decreasing cost");
  par_level[8].push_back("increasing bound range");
  par_level[8].push_back("increasing cost per coefficient count");
  assert( par_level[8].size() == par_levels[8] );

  // MIP branching direction
  par_level[9].push_back("down");
  par_level[9].push_back("automatic");
  par_level[9].push_back("up");
  assert( par_level[9].size() == par_levels[9] );
 */
 assert( par_levels.size() == num_parameters );

  /////////////////////////////////////
  // Now do any initialization required by the program.

  // Initialize the environment
  int status;
  env = CPXopenCPLEX (&status);
  if (env == 0)  
    {
      cout << "Error opening CPLEX environment.  Probably a license issue."
	   << endl;
      exit(1);
    }

  instance = CPXcreateprob (env, &status, "STOP");
  // check status??
}

// This function does any necessary initialization of the metric.
void 
STOP_interface::initialize_metric( void )
{
  // Our metric is total time.  Reset to zero before running the 
  // set of instances.
  metric = 0.0;
}

// Report the metric value to the main program
double 
STOP_interface::get_metric( void )
{
  return metric;
}


// This function sets the parameters based on a vector of integers.
int 
STOP_interface::set_parameters ( vector<int> setting )
{
  int status; 

  // Go back to defaults
  status = CPXsetdefaults (env);

  // Set each of the parameters
  status = CPXsetintparam (env, CPX_PARAM_MIPEMPHASIS, setting[0]);

  status = CPXsetintparam (env, CPX_PARAM_NODESEL, setting[1]+1);

  int temp;
  switch (setting[1]) 
    {
    case 0:  temp = 0; break;
    case 1:  temp = 2; break;
    case 2:  temp = 3; break;
    }
  status = CPXsetintparam (env, CPX_PARAM_VARSEL, temp);

  status = CPXsetintparam (env, CPX_PARAM_DIVETYPE, setting[3]);

  status = CPXsetintparam (env, CPX_PARAM_FRACCUTS, setting[4]-1);

  status = CPXsetintparam (env, CPX_PARAM_MIRCUTS, setting[5]-1);

/*
  status = CPXsetintparam (env, CPX_PARAM_DISJCUTS, setting[6]-1); 

  status = CPXsetintparam (env, CPX_PARAM_CLIQUES , setting[7]-1);

  status = CPXsetintparam (env, CPX_PARAM_PRESLVND, setting[8]-1);

  status = CPXsetintparam (env, CPX_PARAM_PROBE, setting[9]-1);*/

//  status = CPXsetintparam (env, CPX_PARAM_MIPORDTYPE setting[8]);

//  status = CPXsetintparam (env, CPX_PARAM_BRDIR, setting[9]-1);

  // Set time limit
  status = CPXsetdblparam (env, CPX_PARAM_TILIM, max_time);

  return 0;
}


//  This function loads an instance based on a string containing 
// the name of the instance.
int 
STOP_interface::load_instance ( string filename )
{
  int status =  CPXreadcopyprob (env, instance, filename.c_str(), NULL);

  return 0;
}


//  This function runs the test program.
int 
STOP_interface::execute_program ( void )
{

  double time_start = time_now();
  int status = CPXmipopt (env, instance);
  double time_stop = time_now();
  double cur_time = time_stop - time_start;

  status = CPXgetstat(env, instance);

  if (status == CPXMIP_OPTIMAL || status == CPXMIP_OPTIMAL_TOL)
    {
      // Add the time for this instance to the metric
      metric += cur_time;
      return 0;
    }
  else
    {
      // Make the metric equal to double the max time
      // (this is an arbitrary choice indicating a really bad run)
      metric += 2.0*max_time;
      cerr << "CPLEX status is " << status << endl;
      return 1;
    }
}


// Functions to return values of constants

int 
STOP_interface::get_num_initial_tests( void )
{
  return num_initial_tests;
}

int 
STOP_interface::get_num_repetitions( void )
{
  return num_repetitions;
}

int 
STOP_interface::get_num_additional_tests( void )
{
  return num_additional_tests;
}

int
STOP_interface::get_max_time( void )
{
  return max_time;
}

void
STOP_interface::set_max_time( int time )
{
  if (time > 0)  max_time = time;
}

int 
STOP_interface::get_num_parameters( void )
{
  return num_parameters;
}

vector<string> 
STOP_interface::get_par_name( void )
{
  return par_name;
}

vector<int> 
STOP_interface::get_par_levels( void )
{
  return par_levels;
}

vector<int> 
STOP_interface::get_default_settings( void )
{
  return default_settings;
}

vector< vector<string> > 
STOP_interface::get_par_level( void )
{
  return par_level;
}

int
STOP_interface::get_num_instances( void )
{
  return test_instances.size();
}

vector<string> 
STOP_interface::get_test_instances( void )
{
  return test_instances;
}

void
STOP_interface::set_test_instances( vector<string> instances)
{
  test_instances = instances;
}

// Destructor
STOP_interface::~STOP_interface( )
{
  // Nothing needs to be done
}

////////////////////////////////////////////////////
//  Other parameters for potential consideration

  // Cuts
  //CPX_PARAM_CLIQUES 
  //CPX_PARAM_COVERS 
  //CPX_PARAM_DISJCUTS 
  //CPX_PARAM_FLOWCOVERS 
  //CPX_PARAM_FLOWPATHS 
  //CPX_PARAM_GUBCOVERS 
  //CPX_PARAM_IMPLBD 

  // CPX_PARAM_BNDSTRENIND 0, -1, 1
  // CPX_PARAM_COEREDIND 2, 1, 0
  // CPX_PARAM_PRESLVND 0, -1, 1
  // CPX_PARAM_PROBE 0, -1, 1
  // CPX_PARAM_SYMMETRY 

 //
//
