//  interface_glpk.hpp
//
//  Copyright 2007 University of Pittsburgh.
//  U of Pittsburgh coding done by Brady Hunsaker and Mustafa Baz.
//
//  This file is part of STOP (Selection Tool for Optimization Parameters).
//
//  The contents of this file are subject to the Common Public License
//  1.0.  (the "License"); you may not use this file except in
//  compliance with the License. You should have received a copy of
//  the Common Public License along with STOP.
//
//  Software distributed under the License is distributed on an "AS
//  IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
//  implied. See the License for the specific language governing
//  rights and limitations under the License.
//
//  Alternatively, the contents of this file may be used under the
//  terms of the GNU General Public License Version 2 or later (the
//  "GPL"), in which case the provisions of the GPL are applicable
//  instead of those above. If you wish to allow use of your version
//  of this file only under the terms of the GPL, and not to allow
//  others to use your version of this file under the terms of the
//  CPL, indicate your decision by deleting the provisions above and
//  replace them with the notice and other provisions required by the
//  GPL. If you do not delete the provisions above, a recipient may
//  use your version of this file under the terms of either the CPL or
//  the GPL.
//
/////////////////////////////////////////////////////////////

// Additional include files go here
extern "C" {
#include "glpk.h"
}

#include <vector>
#include <string>
#include <cassert>

using namespace std;

#define BIG_DOUBLE 1.0e10;

// function in STOP for timing
double time_now( void );


class STOP_interface
{

public:

  // Initialization is all done in the default constructor.
  STOP_interface ();
  ~STOP_interface ();

  // Metric for testing
  void initialize_metric( void );
  double get_metric( void );

  //  Function to set the parameters
  int set_parameters ( vector<int> setting );
  //  Function to load an instance
  int load_instance ( string filename );
  //  Function to call the program
  int execute_program ( void );

  // Find out the values of constants
  int get_num_initial_tests( void );
  int get_num_repetitions( void );
  int get_num_additional_tests( void );
  int get_num_parameters( void );
  vector<string> get_par_name( void );
  vector<int> get_par_levels( void );
  vector<int> get_default_settings( void );
  vector< vector<string> > get_par_level( void );

  // Max time for each trial
  void set_max_time( int );
  int get_max_time( void );

  // Test instances
  void set_test_instances( vector<string> );
  vector<string> get_test_instances( void );
  int get_num_instances( void );

private:
  ////////////////////////////////////////////////////
  // Any additional variables that are necessary should be defined here.

  // We need a pointer to the instance
  LPX * lp;
  vector<int> current_setting;

  ////////////////////////////////////////////////////
  // Standard variables.  These are all set in the constructor.
  int num_initial_tests;
  int num_additional_tests;
  int num_repetitions;
  int max_time;
  int num_parameters;
  vector<string> par_name;
  vector<int> par_levels;
  vector<int> default_settings;
  vector< vector<string> > par_level;
  //  int num_instances;
  vector<string> test_instances;

  // The metric for comparing settings
  double metric;

}; // end of class



////////////////////////////////////////////////////
// Function definitions

// The default constructor does two things:
//   - sets constant values
//   - does any initialization that may be required by the program

STOP_interface::STOP_interface ()
{
  // Set constant values
  // NOTE: these may be overridden at the command line
  num_initial_tests = 16;
  num_additional_tests = 16;
  num_repetitions = 1;

  max_time = 3600; // 10 minutes


  // Specify test instances here. 
  // Later, may make this specified from the command line 
//  test_instances.push_back ("p0033.mps");
//  test_instances.push_back ("p0201.mps");
 // test_instances.push_back ("p0282.mps");
 //  test_instances.push_back ("misc03.mps");
 //  test_instances.push_back ("misc06.mps");
 //  test_instances.push_back ("misc07.mps");
//   test_instances.push_back ("air03.mps");
  //  test_instances.push_back ("air04.mps");
//CIRCUIT
  // test_instances.push_back ("sobel_fewer_variables_8-1.mps");
  //  test_instances.push_back ("laplace_fewer_variables_8-1.mps");
 //   test_instances.push_back ("adpcm_decoder_fewer_variables_8-1.mps");


//  test_instances.push_back ("airland4_R2.mps");
//  test_instances.push_back ("airland5_R2.mps");
//  test_instances.push_back ("airland5_R1.mps");

// num_instances = test_instances.size();

  // Give a name to each parameter and store the number of levels each has
  par_name.push_back ( "SCALING" );
  par_levels.push_back(4);
  default_settings.push_back(1);
  par_name.push_back ("PRICING");
  par_levels.push_back(2);
  default_settings.push_back(1);
  par_name.push_back ("BRANCHING");
  par_levels.push_back(4);
  default_settings.push_back(2);
  par_name.push_back ("BACKTRACK");
  par_levels.push_back(4);
  default_settings.push_back(2);
  par_name.push_back ("LP PRESOLVE");
  par_levels.push_back(2);
  default_settings.push_back(0);
  par_name.push_back ( "METHOD" );
  par_levels.push_back(3);
  default_settings.push_back(1);

  num_parameters = par_name.size();
  assert( par_levels.size() == num_parameters );

  // For each parameter, give a name to each setting 

  par_level.resize( num_parameters );

  par_level[0].push_back("none");
  par_level[0].push_back("equilibration");
  par_level[0].push_back("geometric mean");
  par_level[0].push_back("geom then equil");
  assert( par_level[0].size() == par_levels[0] );

  par_level[1].push_back("largest coeff");
  par_level[1].push_back("steepest edge");
  assert( par_level[1].size() == par_levels[1] );

  par_level[2].push_back("first var");
  par_level[2].push_back("last var");
  par_level[2].push_back("driebeck-tomlin");
  par_level[2].push_back("most fractional");
  assert( par_level[2].size() == par_levels[2] );

  par_level[3].push_back("depth first");
  par_level[3].push_back("breadth first");
  par_level[3].push_back("best projection");
  par_level[3].push_back("best bound");
  assert( par_level[3].size() == par_levels[3] );

  par_level[4].push_back("off");
  par_level[4].push_back("on");
  assert( par_level[4].size() == par_levels[4] );

  par_level[5].push_back("old b&b");
  par_level[5].push_back("new b&b");
  par_level[5].push_back("new b&b with cuts");
  assert( par_level[5].size() == par_levels[5] );

  /////////////////////////////////////
  // Now do any initialization required by the program.

  // Nothing necessary for GLPK
}

// This function does any necessary initialization of the metric.
void 
STOP_interface::initialize_metric( void )
{
  // Our metric is total time.  Reset to zero before running the 
  // set of instances.
  metric = 0.0;
}

// Report the metric value to the main program
double 
STOP_interface::get_metric( void )
{
  return metric;
}


// This function sets the parameters based on a vector of integers.
int 
STOP_interface::set_parameters ( vector<int> setting )
{
  // Unfortunately, GLPK can't change these settings until the problem 
  // is loaded, so just copy the setting into a global variable and do it
  // later.
  current_setting = setting;

  return 0;
}


//  This function loads an instance based on a string containing 
// the name of the instance.
int 
STOP_interface::load_instance ( string filename )
{
  lp = lpx_read_freemps( const_cast<char *>(filename.c_str() ) );

  // Now set the settings.
  lpx_reset_parms( lp );

  // Check scaling
  lpx_set_int_parm(lp, LPX_K_SCALE, current_setting[0]);

  // Check pricing
  lpx_set_int_parm(lp, LPX_K_PRICE, current_setting[1]);

  // Check branching
  lpx_set_int_parm(lp, LPX_K_BRANCH, current_setting[2]);

  // Check backtrack method
  lpx_set_int_parm(lp, LPX_K_BTRACK, current_setting[3]);

  // Check presolve
  lpx_set_int_parm(lp, LPX_K_PRESOL, current_setting[4]);


  // Set time limit
  lpx_set_real_parm(lp, LPX_K_TMLIM, max_time);

  // Set msg level
  lpx_set_int_parm(lp, LPX_K_MSGLEV, 0);

  return 0;
}


//  This function runs the test program.
int 
STOP_interface::execute_program ( void )
{
  double time_start = time_now();

  int ret = lpx_simplex(lp);

  if (current_setting[5] == 0)
    ret = lpx_integer(lp);
  else
    {
      if (current_setting[5] == 2)
	// turn on cuts
	lpx_set_int_parm(lp, LPX_K_USECUTS, 1);
      else  // turn off cuts
	lpx_set_int_parm(lp, LPX_K_USECUTS, 0);

      // In either case, use the new routine
      ret = lpx_intopt(lp);
    }
  double time_stop = time_now();
  double cur_time = time_stop - time_start;
  
  if (ret == LPX_E_OK)
    {
      // Add the time for this instance to the metric
      metric += cur_time;
      return 0;
    }
  else
    {
      // Make the metric equal to double the max time
      // (this is an arbitrary choice indicating a really bad run)
      metric += 2.0*max_time;
      cerr << "GLPK status is " << ret << endl;
      return 1;
    }
}


// Functions to return values of constants

int 
STOP_interface::get_num_initial_tests( void )
{
  return num_initial_tests;
}

int 
STOP_interface::get_num_additional_tests( void )
{
  return num_additional_tests;
}

int 
STOP_interface::get_num_repetitions( void )
{
  return num_repetitions;
}

int
STOP_interface::get_max_time( void )
{
  return max_time;
}

void
STOP_interface::set_max_time( int time )
{
  if (time > 0)  max_time = time;
}

int 
STOP_interface::get_num_parameters( void )
{
  return num_parameters;
}

vector<string> 
STOP_interface::get_par_name( void )
{
  return par_name;
}

vector<int> 
STOP_interface::get_par_levels( void )
{
  return par_levels;
}

vector< vector<string> > 
STOP_interface::get_par_level( void )
{
  return par_level;
}

vector<int> 
STOP_interface::get_default_settings( void )
{
  return default_settings;
}

int
STOP_interface::get_num_instances( void )
{
  return test_instances.size();
}

vector<string> 
STOP_interface::get_test_instances( void )
{
  return test_instances;
}

void
STOP_interface::set_test_instances( vector<string> instances)
{
  test_instances = instances;
}

// Destructor
STOP_interface::~STOP_interface( )
{
  // Nothing needs to be done
}
